#!/usr/bin/env python
# coding: utf-8

# #### phi2_fvfm_fimg_analysis
# Authors: Francesco Garassino and Sofia Bengoa Luoni  
# Contact: francesco.garassino@wur.nl  
# Last modified: 15/05/2023

# In[1]:


# imports
import glob
import matplotlib.pyplot as plt
import numpy as np
import os
import sys
import time
import warnings
from plantcv import plantcv as pcv


# In[2]:


warnings.filterwarnings("ignore")


# In[21]:


def open_fimg(fimg_path):
    """
    import an image in .fimg format as a numpy array
    """
    with open(fimg_path) as f:
        numpy_data = np.fromfile(f, np.dtype("float32")) # read pixels as 32-bit floating point numbers
    img = numpy_data[2:1392642] #Delete the first 2 characters (these characters contain information about the image size)
    img = np.reshape(img, newshape=(1024, 1360)) # reshape the image to its size
    return img
    
def img_mode(img_array):
    """
    given an image array, will calculate the mode of a 200-bins histogram
    """
    nbins = 200
    binwidth = (np.nanmax(img_array) - np.nanmin(img_array))/nbins
    bins = np.arange(np.nanmin(img_array), np.nanmax(img_array), binwidth)
    h, _ = np.histogram(img_array, bins)
    mode = bins[h.argmax()]
    return mode
    
def mask_img(img_array, param):
    """
    given an array encoding an image, return a mask corresponding to the plant at the center of the image
    """
    img_norm = (img_array - np.min(img_array)) / (np.max(img_array) - np.min(img_array)) # normalize image
    #img_norm_na = np.where(img_norm < 0.1, np.nan, img_norm) # this is needed to remove as much background as possible
    #print(img_mode(img_norm))
    
    ymid = img_norm.shape[0]/2 # center coordinate on the y axis
    xmid = img_norm.shape[1]/2 # center coordinate on the x axis
    img_norm_crop = img_norm[int((ymid-ymid/2)):int((ymid+ymid/2)), int((xmid-xmid/2)):int((xmid+xmid/2))] # crop the image by 50%, centered on the middle
    img_norm_crop_na = np.where(img_norm_crop < 0.09, np.nan, img_norm_crop) # pixels with a normalised value lower than 0.09 (zeros, essentially) will get removed
    #if (np.nanmin(img_norm_crop_na)) != (np.nanmax(img_norm_crop_na)):
        #print(img_mode(img_norm_crop_na))
        
    # if all pixels in the cropped image are NAs (cropped image is empty), the cropped image contains a single value (i.e. is empty), or the mode of pixels is really low (ie - lots of noise)
    if np.isnan(img_norm_crop_na).all() or np.nanmin(img_norm_crop_na) == np.nanmax(img_norm_crop_na): #or img_mode(img_norm_crop_na) < 0.2:
        return np.zeros_like(img_norm), np.nan # return empty stuff
    
    else:
        # create a binary, 8-bit image to feed into plantcv. Binary makes object detection and masking so much quicker!
        if param == 'fvfm':
            img_binary = np.where(img_norm >= 0.2, 1, 0).astype('uint8') # so eveything above the intensity threshold gets a 1, everything below gets a zero
        elif param == 'phiPSII':
            img_binary = np.where(img_norm >= 0.2, 1, 0).astype('uint8') # so eveything above the intensity threshold gets a 1, everything below gets a zero
        else:
            sys.exit('fluorescence parameter not recognised!')

        #plantCV
        objects, obj_hierarchy = pcv.find_objects(img_binary, img_binary)
        img_dims = img_binary.shape
        roi, roi_hierarchy = pcv.roi.rectangle(img = img_binary, x = (img_dims[1]/2)-150, y = (img_dims[0]/2)-150, h = 300, w = 300)
        roi_objects, hierarchy, kept_mask, obj_area = pcv.roi_objects(img = img_binary,
                                                                      roi_contour = roi,
                                                                      roi_hierarchy = roi_hierarchy,
                                                                      object_contour = objects,
                                                                      obj_hierarchy = obj_hierarchy,
                                                                      roi_type="partial")
        mask_filled = pcv.fill(bin_img=kept_mask, size=20)
        mask_32bit = (mask_filled/255).astype('float32') # NB dividing by 255 will give us a binary mask with zeros and ones
        return mask_32bit, obj_area

def shortenImgName(img_path):
    """
    returns shortened name of image to be used for outputting
    
    :img_path: str, containing a full path to a single f-image
    :return: img_name_short, str, contains expID, roundID, trayID for a specific plant measured at a specific time - something like 35-66-Tray_Circle_101
    """
    img_name_short = img_path.split("\\")[-1].split("-FC1")[0]
    return img_name_short

def calc_fvfm(path_fo, path_fm, output_folder = os.getcwd(), hist = False, fig = False, cbar_both = True):
    """
    given the paths to fp and fmp images, will create an image array containing calculated phi2 values, calculate summary statistics and optionally export images
    """
    # tracking time
    #start_time = time.process_time()
    # prepare for saving images
    images_name = shortenImgName(path_fo) # retrieve name for outputs
    roundnr = images_name.split('-')[-2]
    tray = images_name.split('-')[-1]
    if not output_folder.endswith('\\'):
        output_folder += '\\'  
        
    fm_img = open_fimg(path_fm)
    mask, fvfm_area = mask_img(fm_img, param = 'fvfm') # create mask
    fm_masked = fm_img*mask # apply mask
    fm_flat = fm_masked.flatten() # convert back to 1D array
    #print('\t\ttime until opening and masking Fm image: {}s'.format(time.process_time() - start_time, "seconds"))
    
    fo_img = open_fimg(path_fo) # open image
    fo_masked = fo_img*mask # apply same mask as before - will make things faster!
    fo_flat = fo_masked.flatten() # convert back to 1D array
    #print('\t\ttime until opening and masking Fo image: {}s'.format(time.process_time() - start_time, "seconds"))
    
    # calculates Fv/Fm, setting NAs to points where fm was set to zero (mask) or fo is higher than fm (errors?)
    fvfm_flat = np.where(fm_flat > 0, np.where(fm_flat > fo_flat, (fm_flat - fo_flat)/fm_flat, np.nan), np.nan)
    
    # retrieve descriptive stats
    fvfm_max = np.nanmax(fvfm_flat)
    fvfm_min = np.nanmin(fvfm_flat)
    fvfm_mean = np.nanmean(fvfm_flat)
    fvfm_median = np.nanmedian(fvfm_flat)  
    #print('\t\ttime until caclulations: {}s'.format(time.process_time() - start_time, "seconds"))
    
    if hist:
        if np.isnan(fvfm_flat).all() and np.isnan(fvfm_flat).all():
            print('\t no data in figure, no histogram plotted')
            fvfm_mode = np.nan
        else:
             # calculating the mode, i.e. the most represented value in the histogram
            nbins = 200
            binwidth = (np.nanmax(fvfm_flat) - np.nanmin(fvfm_flat))/nbins
            bins = np.arange(np.nanmin(fvfm_flat), np.nanmax(fvfm_flat), binwidth)
            h, _ = np.histogram(fvfm_flat, bins)
            fvfm_mode = bins[h.argmax()]
            # will save an histogram of values 
            plt.hist(fvfm_flat, color='w', ec='grey', bins=bins, range=[0.2, 1])
            plt.axvline(x=fvfm_mode, color='blue', linestyle='-')
            plt.axvline(x=fvfm_mean, color='red', linestyle='--')
            plt.savefig('{}{}-FvFm_hist.jpg'.format(output_folder, images_name), dpi=300)
            plt.clf()
        
    if fig:
        # will save an image of Fv/Fm
        fvfm = np.reshape(fvfm_flat, newshape=(1024, 1360))
        fvfm = np.where(np.isnan(fvfm), 0, fvfm)
        plt.imshow(fvfm, vmin = 0.2, vmax = 1, cmap="nipy_spectral")
        plt.colorbar(fraction=0.033)
        plt.savefig('{}{}-FvFm.jpg'.format(output_folder, images_name), dpi=300)
        plt.clf()
        
        if cbar_both:
            plt.imshow(fvfm, vmin = 0.2, vmax = 1, cmap="nipy_spectral")
            plt.savefig('{}{}-FvFm_noscale.jpg'.format(output_folder, images_name), dpi=300)
            plt.clf()
    
    #print('\t\tcumulative time: {}s'.format(time.process_time() - start_time, "seconds"))
    return [roundnr, tray, fvfm_area, fvfm_max, fvfm_min, fvfm_mean, fvfm_median, fvfm_mode]

def calc_phi2(path_fp, path_fmp, output_folder = os.getcwd(), hist = False, fig = False, cbar_both = False):
    """
    given the paths to fp and fmp images, will create an image array containing calculated phi2 values, calculate summary statistics and optionally export images
    """
    # tracking time
    #start_time = time.process_time()
    # prepare for saving images
    images_name = shortenImgName(path_fp) # retrieve name for outputs
    roundnr = images_name.split('-')[-2]
    tray = images_name.split('-')[-1]
    if not output_folder.endswith('\\'):
        output_folder += '\\'    
    
    fmp_img = open_fimg(path_fmp)
    mask, phi2_area = mask_img(fmp_img, param = 'phiPSII') # create mask
    
    if np.isnan(phi2_area):
        print('\t\tsignal-to-background noise ratio too small, will not process the image')
        return []
    else:
        fmp_masked = fmp_img*mask # apply mask
        fmp_flat = fmp_masked.flatten() # convert back to 1D array
        #print('\t\ttime until opening and masking of Fmp image: {}s'.format(time.process_time() - start_time, "seconds"))

        fp_img = open_fimg(path_fp) # open image
        fp_masked = fp_img*mask # apply apply same mask as before - will make things faster!
        fp_flat = fp_masked.flatten() # convert back to 1D array
        #print('\t\ttime until opening and masking of Fp image: {}s'.format(time.process_time() - start_time, "seconds"))

        # calculates phi2, setting NAs to points where fmp was set to zero (mask) or fp is higher than fmp (errors?)
        phi2_flat = np.where(fmp_flat > 0, np.where(fmp_flat > fp_flat, (fmp_flat - fp_flat)/fmp_flat, np.nan), np.nan)

        # retrieve descriptive stats
        phi2_max = np.nanmax(phi2_flat)
        phi2_min = np.nanmin(phi2_flat)
        phi2_mean = np.nanmean(phi2_flat)
        phi2_median = np.nanmedian(phi2_flat)

        #print('\t\ttime until calculations: {}s'.format(time.process_time() - start_time, "seconds"))
        if hist:
            if np.isnan(phi2_flat).all() and np.isnan(phi2_flat).all():
                print('\tno data in figure, no histogram plotted')
                phi2_mode = np.nan
            else:      
                # calculating the mode, i.e. the most represented value in the histogram
                nbins = 200
                binwidth = (np.nanmax(phi2_flat) - np.nanmin(phi2_flat))/nbins
                bins = np.arange(np.nanmin(phi2_flat), np.nanmax(phi2_flat), binwidth)
                h, _ = np.histogram(phi2_flat, bins)
                phi2_mode = bins[h.argmax()]
                # will save an histogram of values 
                plt.hist(phi2_flat, color='w', ec='grey', bins=200, range=[0, 0.8])
                plt.axvline(x=phi2_mode, color='blue', linestyle='-')
                plt.axvline(x=phi2_mean, color='red', linestyle='--')
                plt.savefig('{}{}-phiPSII_hist.jpg'.format(output_folder, images_name), dpi=300)
                plt.clf()

        if fig:
            # will save an image of Phi2
            phi2 = np.reshape(phi2_flat, newshape=(1024, 1360))
            phi2 = np.where(np.isnan(phi2), 0, phi2)
            plt.imshow(phi2, vmin = 0, vmax = 0.8, cmap="nipy_spectral")
            plt.colorbar(fraction=0.033)
            plt.savefig('{}{}-phiPSII.jpg'.format(output_folder, images_name), dpi=300)
            plt.clf()

            if cbar_both:
                plt.imshow(phi2, vmin = 0, vmax = 0.8, cmap="nipy_spectral")
                plt.savefig('{}{}-phiPSII_noscale.jpg'.format(output_folder, images_name), dpi=300)
                plt.clf()

        #print('\t\tcumulative time: {}s'.format(time.process_time() - start_time, "seconds"))
        return [roundnr, tray, phi2_area, phi2_max, phi2_min, phi2_mean, phi2_median, phi2_mode]

print('functions OK')


# In[16]:


data_path = "D:\\NV Spring 2022 - 2nd download\\FC1\\Analysis\\ParametersImages\\"
out_path = "C:\\Users\\garas001\\OneDrive - Wageningen University & Research\\PhD\\_Year 4\\Chapter 4 - Natural Variation B8\\1_Fluorescence analysis\\" + "single_images_09_08\\"
# dev
#out_path = "C:\\Users\\garas001\\OneDrive - Wageningen University & Research\\PhD\\_Year 4\\Chapter 4 - Natural Variation B8\\Analysis 2023\\" + "test2\\"


# In[24]:


# make sure we have the correct output directory
if not os.path.exists(out_path):
    os.makedirs(out_path)


# In[18]:


# create tray numbers in the right format
tray_nrs  = range(1,181,1)
tray_nrs = [str(item).zfill(3) for item in tray_nrs]

# remove trays in which plants did not grow
bad_trays = [9, 21, 33, 46, 52, 59, 66, 80, 82, 89, 94, 95, 99, 100, 113, 117, 127, 128, 129, 142, 144, 148, 154, 155, 167, 169, 173, 174]
bad_trays_nrs = [str(item).zfill(3) for item in bad_trays]
tray_nrs = [x for x in tray_nrs if x not in bad_trays_nrs]


# In[19]:


print('total number of surviving plants: {}'.format(len(tray_nrs)))


# In[25]:


# FvFm operations
fvfm = True

if fvfm:

    print('\nWorking on Fv/Fm...')
    fvfm_stats_list = [['round_nr', 'tray_id', 'fvfm_area', 'fvfm_max', 'fvfm_min', 'fvfm_mean', 'fvfm_median', 'fvfm_mode']]
    #tray_nrs = [101]

    for nr in tray_nrs:
        fo_fimgs = sorted(glob.glob('{}*{}*-Fo.fimg'.format(data_path, nr)), key=os.path.getmtime)
        selected_fo_fimgs = []
        for path in fo_fimgs:
            if int(path.split("\\35-")[-1].split("-")[0]) >= 1: # we're selecting from round 22 as 21 is problematic. This gets rid of most images with little plant signal which appear to be a problem for PlantCV
                selected_fo_fimgs.append(path)

        fm_fimgs = sorted(glob.glob('{}*{}*-Fm.fimg'.format(data_path, nr)), key=os.path.getmtime)
        selected_fm_fimgs = []
        for path in fm_fimgs:
            if int(path.split("\\35-")[-1].split("-")[0]) >= 1:
                selected_fm_fimgs.append(path)

        if os.path.exists(out_path + '35-92-Tray_Circle_{}-FvFm.jpg'.format(nr)):
            print ('Images from tray {} already processed'.format(nr))
            continue
        else:
            start_time = time.process_time()
            print('\nProcessing images from tray {}'.format(nr))

            if len(selected_fo_fimgs) != len(selected_fm_fimgs):
                sys.exit('numbers of Fp and Fmp images do not match') # "brake" in case something funny is going on with input images
            elif len(selected_fo_fimgs) == 0:
                 sys.exit('no images found!')

            for n in range(0, len(selected_fo_fimgs)):
                print('\t ROUND {}'.format(selected_fo_fimgs[n].split('\\35-')[-1].split('-')[0]))
                fvfm_stats = calc_fvfm(selected_fo_fimgs[n], selected_fm_fimgs[n], output_folder = out_path, hist = True, fig = True, cbar_both = True)
                fvfm_stats_list += [fvfm_stats]

            print('\t\texecution time for tray {}: {}s'.format(nr, time.process_time() - start_time, "seconds"))

    base_folder = '\\'.join(out_path.split('\\')[:-2]) + '\\'        
    fvfm_array = np.asarray(fvfm_stats_list)
    np.savetxt('{}fvfm_stats.csv'.format(base_folder), fvfm_array, delimiter=",", fmt="%s")


# In[26]:


# Phi2 operations

phiPSII = True

if phiPSII:
    # dev
    #tray_nrs = ['001']

    print('\nProcessing phiPSII...')
    t0 = time.process_time()
    phi2_stats_list = [['round_nr', 'tray_id', 'phi2_area', 'phi2_max', 'phi2_min', 'phi2_mean', 'phi2_median', 'phi2_mode']]

    for nr in tray_nrs:
        fp_fimgs = sorted(glob.glob('{}*{}*Fp.fimg'.format(data_path, nr)), key=os.path.getmtime)
        selected_fp_fimgs = []
        for path in fp_fimgs:
            selected_fp_fimgs.append(path)

        fmp_fimgs = sorted(glob.glob('{}*{}*Fmp.fimg'.format(data_path, nr)), key=os.path.getmtime)
        selected_fmp_fimgs = []
        for path in fmp_fimgs:
            if int(path.split("\\35-")[-1].split("-")[0]) >= 1:
                selected_fmp_fimgs.append(path)

        if os.path.exists(out_path + '35-86-Tray_Circle_{}-phiPSII.jpg'.format(nr)):
            print ('Images from tray {} already processed'.format(nr))
            continue
        else:
            start_time = time.process_time()
            print('\nProcessing images from tray {}'.format(nr))

            if len(selected_fp_fimgs) != len(selected_fmp_fimgs):
                sys.exit('numbers of Fp and Fmp images do not match') # "brake" in case something funny is going on with input images

            for n in range(0, len(selected_fp_fimgs)):
                # # exclude round 21 as it's problematic
                # if '35-21' in selected_fp_fimgs[round]:
                #     continue
                round_str = selected_fp_fimgs[n].split('35-')[-1].split('-')[0]
                print('\t ROUND {}'.format(round_str))
                if round_str == '21':
                      print('\t\tskipping this one...') # round 21 had some error
                else:
                    phi2_stats = calc_phi2(selected_fp_fimgs[n], selected_fmp_fimgs[n], output_folder = out_path, hist = True, fig = True, cbar_both = True)
                    if len(phi2_stats) == 0 :
                        pass
                    else:
                        phi2_stats_list += [phi2_stats]

            print('\t\texecution time for tray {}: {}s'.format(nr, time.process_time() - start_time, "seconds"))

    base_folder = '\\'.join(out_path.split('\\')[:-2]) + '\\'
    phi2_array = np.asarray(phi2_stats_list)
    np.savetxt('{}phi2_stats.csv'.format(base_folder), phi2_array, delimiter=",", fmt="%s")

    print('total time: {}'.format(time.process_time() - t0, "minutes"))

